/*
 * @(#)FibonacciSamples.java  1.0  4 March 2002
 *
 * Fibonacci Samples
 *
 * Copyright 2001,2002
 * College of Computer Science
 * Northeastern University
 * Boston, MA  02115
 */

import edu.neu.ccs.console.*;

/**
 * This class collects various approaches to computing
 * Fibonacci numbers.
 *
 * The class is ConsoleAware in case the student wants
 * to print information to monitor the progress of the
 * algorithms.
 */
public class FibonacciSamples implements ConsoleAware
{
    /**
     * SimpleFibonacci implements Fibonacci in the simple
     * manner suggested by the mathematical definition
     * but is in fact extraordinarily inefficient.  Hence
     * this function is a terrible model for recursion.
     */
    public static double SimpleFibonacci(int n) {
        // feedback
        console.out.println("SimpleFibonacci(" + n + ")");
        
        if (n < 0)
            return 0;
        
        if (n < 2)
            return 1;
        
        return SimpleFibonacci(n-2) + SimpleFibonacci(n-1);
    }
    
    /**
     * LoopFibonacci avoids recursion altogether and does
     * the Fibonacci computation in a loop.
     */
    public static double LoopFibonacci(int n) {
        // feedback
        console.out.println("LoopFibonacci(" + n + ")");
        
        if (n < 0)
            return 0;
        
        if (n < 2)
            return 1;
        
        // loop setup
        int k = 1;          // current known index of Fibonacci
        double a = 1;       // Fibonacci at k-1
        double b = 1;       // Fibonacci at k
        double c = 2;       // sum variable
        
        // compute Fibonacci at k up to k == n
        while (k < n) {
            // feedback
            console.out.println
                ("Loop For " + k + ", " + a + ", " + b);
            
            c = a + b;
            a = b;
            b = c;
            k++;
        }
        
        // feedback
        console.out.println
            ("At End   " + k + ", " + a + ", " + b);
        
        // return final Fibonacci
        return b;
    }
    
    /**
     * SubtleFibonacci uses recursion in an efficient manner by
     * the technique of passing data downwards using the extra
     * tail recursive function HelperFibonacci.
     */
    public static double SubtleFibonacci(int n) {
        // feedback
        console.out.println("SubtleFibonacci(" + n + ")");
        
        if (n < 0)
            return 0;
        
        if (n < 2)
            return 1;
        
        // This call to HelperFibonacci is based on:
        //     We know Fibonacci at 1 (k)
        //     Fibonacci at 0 is 1 (a)
        //     Fibonacci at 1 is 1 (b)
        return HelperFibonacci(n, 1, 1, 1);
    }
    
    /**
     * The tail recursive HelperFibonacci for SubtleFibonacci.
     * This function expects that k <= n.
     *
     * @param n we want Fibonacci at n
     * @param k we know Fibonacci at k
     * @param a the value of Fibonacci at k-1
     * @param b the value of Fibonacci at k
     */
    protected static double HelperFibonacci(int n, int k, double a, double b) {
        // feedback
        console.out.println
            ("HelperFibonacci(" + n + ", " + k + ", " + a + ", " + b + ")");
        
        if (n <= k)
            return b;
        
        return HelperFibonacci(n, k+1, b, a+b);
    }
    
    /**
     * <P>FancyFibonacci uses a mathmatical formula for Fibonacci(n)
     * that involves the n-th powers of the roots of the equation:
     * X^2 - X - 1 = 0.
     *
     * <P>The roots of this equation are:
     * <UL>
     *   <LI> gamma = (1 + sqrt5) / 2
     *   <LI> delta = (1 - sqrt5) / 2
     * </UL>
     *
     * <P>Using these roots, the formula for Fibonacci(n) is:
     * <UL>
     *   <LI>Fibonacci(n) = a * gamma^n + b * delta^n
     * </UL>
     * 
     * <P>The coefficients a, b are given as:
     * <UL>
     *   <LI> a = (1 - delta) / (gamma - delta) = + gamma / sqrt5
     *   <LI> b = (1 - gamma) / (delta - gamma) = - delta / sqrt5
     * </UL>
     *
     * <P>These coefficients are chosen so that the formula is correct
     * for n = 0, 1 and then the mathematics shows that the formula is
     * correct for all n >= 2.
     */
    public static double FancyFibonacci(int n) {
        // feedback
        console.out.println("FancyFibonacci(" + n + ")");
        
        if (n < 0)
            return 0;
        
        // compute the roots gamma and delta of the equation: X^2 - X - 1 = 0
        double sqrt5 = Math.sqrt(5);
        
        double gamma = (1 + sqrt5) / 2;
        double delta = (1 - sqrt5) / 2;
        
        // compute coefficients for Fibonacci
        double a = + gamma / sqrt5;
        double b = - delta / sqrt5;
        
        // Fibonacci(n) = a * gamma^n + b * delta^n
        return Math.round(a * power(gamma, n) + b * power(delta, n));
    }
    
    /** Returns the given x raised to the given integer power n. */
    public static double power(double x, int n) {
        if (n < 0) {
            n = - n;
            x = 1/x;
        }
        
        if (n == 0)
            return 1;
        
        if (n == 1)
            return x;
        
        double y = power(x, n / 2);
        
        if ((n % 2) == 0)
            return y * y;
        else
            return y * y * x;
    }
    
}
